/*******************************************************************\

Module:

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#include <stdio.h>

#include <string>
#include <fstream>

#include <dblib/os_dep_file.h>
#include <dblib/os_dep_process.h>
#include <dblib/str_getline.h>

#include "lock.h"
#include "logstr.h"
#include "client.h"

//
// Locking Services
// 
// The types/name combinations used should be unique
//
// Todo: This locking does not work over NFS
//       This locking probably does not work with Win32
//       Automatic stale lock detection has a race condition
//

/*******************************************************************\

Function: mklockname

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

static void mklockname(std::string &filename,
  const std::string &type, const std::string &name)
 {
  std::string lockdir=client.config["lockdir"];

  if(lockdir=="")
   {
    logstr << "configuration error: lockdir not set";
    logstr.log("mklockname");

    throw client_excpt("configuration error: lockdir not set");
   }

  std::string tmpname=name;
  
  for(unsigned i=0; i<tmpname.size(); i++)
    if(tmpname[i]==DIRCHAR[0])
      tmpname[i]='_';

  tmpname=type+"."+tmpname;

  filename=build_path(lockdir, tmpname);
 }

/*******************************************************************\

Function: stale_lock

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool stale_lock(const std::string &filename)
 {
  std::string line1, line2, line3;
  std::ifstream infile(filename.c_str());

  str_getline_crlf(infile, line1);
  str_getline_crlf(infile, line2);
  str_getline_crlf(infile, line3);

  logstr << "warning: stale lock " << filename
         << " detected: " << line2 << "@" << line1 
         << " module " << line3;
  logstr.log("stale_lock");

  char tempstr[200];
  gethostname(tempstr, 199);

  if(line1==tempstr) // local machine?
   {
    #ifndef WIN32

    if(kill(atoi(line2.c_str()), 0)!=0 && errno==ESRCH) // process dead
     {
      logstr << line2 << "@" << line1 << " dead, removing lock";
      logstr.log("stale_lock");

      unlink(filename.c_str()); // race: another process might do the same
      return FALSE;
     }
    else // process not dead
     {
      logstr << "stale_lock" << line2 << "@" << line1
             << " is blocking resource " << filename;
      logstr.log("stale_lock");
      return FALSE;
     }

    #endif
   }
  else // remote machine
   {
    logstr << "warning: removing remote lock (unsafe)";
    logstr.log("stale_lock");
    unlink(filename.c_str());
    return FALSE;
   }  

  return TRUE;
 }

/*******************************************************************\

Function: lock

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void lock(const std::string &type, const std::string &name,
          const std::string &module)
 {
  std::string filename;
  int handle, attempt=0;

  mklockname(filename, type, name);

  while(TRUE)
   {
    attempt++;

    handle=open(filename.c_str(), O_WRONLY|O_CREAT|O_EXCL|O_BINARY, 0600);

    if(handle<0) 
     {
      if(attempt>=2) 
       {
        #ifdef WIN32
        Sleep(1);
        #else
        sleep(1);
        #endif
       }
      else
       {
        #ifdef WIN32
        Sleep(1);
        #else
        #ifdef OS_WIN32
        sleep(1);
        #else
        sched_yield();
        #endif
        #endif
       }

      if(attempt>20)
       {
        if(stale_lock(filename))
          throw client_excpt("failed to acquire lock "+filename);

        attempt=0;
       }
     }
    else
      break;
   }

  char tempstr[100];

  gethostname(tempstr, 99);
  write(handle, tempstr, strlen(tempstr));

  #ifdef WIN32
  snprintf(tempstr, 99, "\n\n");
  #else
  snprintf(tempstr, 99, "\n%d\n", getpid());
  #endif
  write(handle, tempstr, strlen(tempstr));

  if(module!="")
   {
    write(handle, module.c_str(), module.size());
    strncpy(tempstr, "\n", 99);
    write(handle, tempstr, strlen(tempstr));
   }

  close(handle);
 }

/*******************************************************************\

Function: unlock

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void unlock(const std::string &type, const std::string &name)
 {
  std::string filename;
  mklockname(filename, type, name);
  unlink(filename.c_str());
 }

